from saga.base_file.File import File


def parse_text_file(file_id, file_name, file_path):
    f = open(file_path, "r")
    file_contents = []
    for line in f.readlines():
        if line.endswith("\n"):
            file_contents.append(line[:len(line) - 1])
        else:
            file_contents.append(line)

    return File(file_id, "text", file_path, file_name, file_contents)


def write_text_file(file):
    f = open(file.file_path, "w+")
    f.write("\n".join(file.file_contents.mixed_data_type))
    f.close()


from saga.base_file.File import File

def parse_binary_file(file_id, file_name, file_path):
    f = open(file_path, "r")
    file_contents = [f.read()]
    f.close()
    return File(file_id, "binary", file_path, file_name, file_contents)

def write_binary_file(file):
    f = open(file.file_path, "wb+")
    f.write(bytes(file.file_contents.mixed_data_type[0], "utf-8"))
    f.close()




import pytest
import os
from saga.base_file.File import File
from saga.file_types.binary_file import parse_binary_file, write_binary_file

def create_binary_file(name, contents):
    if os.path.isfile(name):
        os.remove(name)
    f = open(name, 'w+b')
    f.write(bytearray(contents, 'utf-8'))
    f.close()

@pytest.fixture()
def setup_binary_files():
    if not os.path.isdir("temp"):
        os.mkdir('temp')
    create_binary_file("temp/binary0", "12345")

def test_create(setup_binary_files):
    binary_file = parse_binary_file("id", "name", "temp/binary0")
    assert binary_file.file_name == "name"
    assert binary_file.file_contents.mixed_data_type[0] == '12345'

def test_create_then_write(setup_binary_files):
    binary_file = parse_binary_file("id", "name", "temp/binary0")
    write_binary_file(binary_file)
    f = open('temp/binary0', 'rb')
    assert f.read() == b'12345'

def test_get_operations_no_changechange(setup_binary_files):
    binary_file0 = parse_binary_file("id", "name", "temp/binary0")
    binary_file1 = parse_binary_file("id", "name", "temp/binary0")
    ops = binary_file0.get_operations(binary_file1)
    for key in ops:
        assert len(ops[key]) == 0

def test_get_operations_change(setup_binary_files):
    binary_file0 = parse_binary_file("id", "name", "temp/binary0")
    create_binary_file("temp/binary0", "54321")
    binary_file1 = parse_binary_file("id", "name", "temp/binary0")
    ops = binary_file0.get_operations(binary_file1)
    assert len(ops["changed"]) == 1
    assert ops["changed"][0] == [0]


from saga.base_file.mixed_data_type.lcs import lcs_with_sim
from saga.base_file.mixed_data_type.diff_utils import dict_removed_paths, dict_inserted_paths, dict_changed_paths

PRIMITIVE = (int, float, bool, str, type(None))

def merge_rec(O, A, B):
    # If one variable is a primitive type, all should be primitive types
    if type(O) in PRIMITIVE or type(A) in PRIMITIVE or type(B) in PRIMITIVE:
        # if a value has been changed in at most one place, then we can merge it
        assert type(O) in PRIMITIVE and type(A) in PRIMITIVE and type(B) in PRIMITIVE
        if O == A and O == B:
            return O
        elif O != A and O == B:
            return A
        elif O == A and O != B:
            return B
        else:
            return None
    
    # All variables must have the same type!
    assert type(O) == type(A) and type(A) == type(B)

    if type(O) == dict:
        o_a_changed = dict_changed_paths(O, A)
        o_a_inserted = dict_inserted_paths(O, A)
        o_a_removed = dict_removed_paths(O, A)

        o_b_changed = dict_changed_paths(O, B)
        o_b_inserted = dict_inserted_paths(O, B)
        o_b_removed = dict_removed_paths(O, B)

        # if we insert something in both, then we have a merge conflict:
        if any(o_a_inserted.intersection(o_b_inserted)):
            # TODO: we could also not make it a merge conflict, if they add the same thing
            return None
        # if we removed something from one, and change it in the other
        if any(o_a_removed.intersection(o_b_changed)) or any(o_b_removed.intersection(o_a_changed)):
            return None

        new_dict = {k: O[k] for k in O}

        # if they were changed in both, we need to merge 
        for changed in o_a_changed.intersection(o_b_changed):
            merged = merge_rec(O[changed], A[changed], B[changed])
            if merged is None:
                return None
            new_dict[changed] = merged

        for removed in o_a_removed.union(o_b_removed):
            del new_dict[removed]
        for a in o_a_changed.union(o_a_inserted).difference(o_b_changed):
            new_dict[a] = A[a]
        for b in o_b_changed.union(o_b_inserted).difference(o_a_changed):
            new_dict[b] = B[b]

        return new_dict

    elif type(O) == list:
        # we run a modified version of the diff3 algorithm in this case
        matchings_A = bump_matching(lcs_with_sim(A, O))
        matchings_B = bump_matching(lcs_with_sim(B, O))
        chunks = get_chunks(O, A, B, matchings_A, matchings_B)

        calculated_ouput = []

        for chunk in chunks:

            ((s_O, e_O), (s_A, e_A), (s_B, e_B)) = chunk

            a_changed = chunk_changed(matchings_A, s_A, e_A, s_O, e_O)
            b_changed = chunk_changed(matchings_B, s_B, e_B, s_O, e_O)

            if a_changed and not b_changed:
                calculated_ouput.extend(chunk_changed_in_A(chunk, A))
            elif b_changed and not a_changed:
                calculated_ouput.extend(chunk_changed_in_B(chunk, B))
            else:
                chunk_o, chunk_a, chunk_b = chunk_conflicting_or_stable(chunk, O, A, B)
                if not contains_primitive(chunk_a) \
                    and not contains_primitive(chunk_b) and not contains_primitive(chunk_o):
                    # There's a "location" of these sublist elements in the final list, and then there's the combining
                    # of the sublists when they are equal... 

                    # TODO: not sure about this condition
                    if len(chunk_a) > len(chunk_o) and len(chunk_b) > len(chunk_o):
                        # ignore nones!

                        overlay = none_overlay(chunk_a, chunk_b)
                        print(f"OVERLAY {overlay}")
                        if overlay is not None:
                            calculated_ouput.extend(overlay)
                            continue
                        else:
                            return None
        
                    for o, a, b in zip(chunk_o, chunk_a, chunk_b):
                        rec_merge = merge_rec(o, a, b)
                        if rec_merge is None:
                            return None
                        calculated_ouput.append(rec_merge)  

                else:
                    if chunk_a != chunk_b:
                        return None

                    calculated_ouput.extend(chunk_a)

        return calculated_ouput
    else:
        raise ValueError("Invalid type given to merge: {}".format(type(O)))

# utilities for diff3

def none_overlay(matrix_a, matrix_b):
    import itertools
    # overlays 2 2-d arrays, where None is treated as a blank space
    overlay = []
    for a, b in itertools.zip_longest(matrix_a, matrix_b):
        if a == None:
            overlay.append(b)
        elif b == None:
            overlay.append(a)
        else:
            overlay.append([])
            idx = len(overlay) - 1
            for x, y in itertools.zip_longest(a, b):
                if x == y:
                    overlay[idx].append(x)
                elif x == None:
                    overlay[idx].append(y)
                elif y == None:
                    overlay[idx].append(x)
                else:
                    return None
    return overlay

def bump_matching(matching):
    new_matchings = []
    for path1, path2, sim in matching:
        path1 = [i + 1 for i in path1]
        path2 = [i + 1 for i in path2]
        new_matchings.append((path1, path2, sim))
    return new_matchings


def get_chunks(O, A, B, matchings_A, matchings_B):
    chunks = []
    idx_A, idx_O, idx_B = 0, 0, 0
    max_idx_A, max_idx_O, max_idx_B = len(A), len(O), len(B)

    i = 1
    # we keep trying to make chunks while we haven't run off the end of any list
    while not (i + idx_A > max_idx_A or i + idx_O > max_idx_O or i + idx_B > max_idx_B):
        # if there is no longer a matching between these indicies
        if not is_matching(matchings_A, idx_A + i, idx_O + i) or not is_matching(matchings_B, idx_B + i, idx_O + i):
            # and i is 1 
            if i == 1:
                # then we find the end index of the unstable chunk (e.g. while it remains stable)
                found = False
                for j in range(idx_O + 1, max_idx_O + 1):
                    last_indexes = is_stable_index(matchings_A, matchings_B, j)
                    if last_indexes:
                        found = True
                        # found last index of unstable chunk
                        (end_A, end_O, end_B) = last_indexes
                        unstable_chunk = ((idx_O + 1, end_O - 1), (idx_A + 1, end_A - 1), (idx_B + 1, end_B - 1))
                        chunks.append(unstable_chunk)
                        idx_A, idx_O, idx_B = end_A - 1, end_O - 1, end_B - 1
                        i = 1
                        break
                # thus, there is no matchings, and so we output one more final chunk
                if not found:
                    unstable_chunk = ((idx_O + 1, max_idx_O), (idx_A + 1, max_idx_A), (idx_B + 1, max_idx_B))
                    chunks.append(unstable_chunk)
                    return chunks
            # otherwise, this is a stable chunk, and we just output it
            else:
                stable_chunk = ((idx_O + 1, idx_O + i - 1), (idx_A + 1, idx_A + i - 1), (idx_B + 1, idx_B + i - 1))
                chunks.append(stable_chunk)
                idx_A, idx_O, idx_B = idx_A + i - 1, idx_O + i - 1, idx_B + i - 1
                i = 1
        else:
            i += 1
    else:
        # we made one final chunk
        last_chunk = ((idx_O + 1, max_idx_O), (idx_A + 1, max_idx_A), (idx_B + 1, max_idx_B))
        chunks.append(last_chunk)
        return chunks

def contains_primitive(l):
    return len(l) > 0 and type(l[0]) in PRIMITIVE    

def chunk_changed_in_A(chunk, A):
    #output A[H], A[H], A[H]
    (_, (s_A, e_A), _) = chunk
    return A[s_A - 1 : e_A]

def chunk_changed_in_B(chunk, B):
    #output B[H], B[H], B[H]
    (_, _, (s_B, e_B)) = chunk
    return B[s_B - 1 : e_B]

def chunk_conflicting_or_stable(chunk, O, A, B):
    #output A[H], O[H], B[H]
    ((s_O, e_O), (s_A, e_A), (s_B, e_B)) = chunk
    return (O[s_O - 1: e_O ], A[s_A - 1: e_A], B[s_B - 1: e_B ])

def chunk_changed(matchings, s, e, s_O, e_O):
    for i in range (s, e + 1):
            if matched(matchings, i) != 1:
                return True

    for i in range (s_O, e_O + 1):
            if matched_O(matchings, i) != 1:
                return True

    return False

# returns true if a matching exists with this index of A or B
def matched(matchings, idx):
    for match in matchings:
        ([i], _, sim) = match
        if (i == idx):
            return sim 
    return 0

# returns true if a matching exists with this index of A or B
def matched_O(matchings, idx):
    for match in matchings:
        (_, [i], sim) = match
        if (i == idx):
            return sim
    return 0

# i is matched to j
def is_matching(matchings, i, j):
    for match in matchings:
        (i_prime, j_prime, similarity) = match
        if i == i_prime[0] and j == j_prime[0]:
            return True
    return False

def is_stable_index(matchings_A, matchings_B, idx_O):
    for match_A in matchings_A:
        if match_A[1][0] == idx_O:
            for match_B in matchings_B:
                if match_B[1][0] == idx_O:
                    # return (idx_A, idx_O, idx_B)
                    return (match_A[0][0], match_A[1][0], match_B[0][0])
    return False 



from saga.base_file.File import File
from bs4 import BeautifulSoup
from zipfile import ZipFile

def char_position(letter):
    return ord(letter) - 65

def get_index(column):
    # for now, we're only gonna do single letter columns
    return char_position(column)

def get_strings(zf):
    try:
        with zf.open('xl/sharedStrings.xml') as stringFile:
            soup = BeautifulSoup(stringFile.read(), 'html5lib')
            unique_strs =  [None] * int(soup.body.sst["count"])
            for i, val in enumerate(soup.body.sst):
                unique_strs[i] = str(val.contents[0].contents[0])
        return unique_strs
    except:
        return []

def get_rows(sheetdata, strings):
    rows = dict()

    # we check if the sheet is empty
    if len(sheetdata) <= 1:
        return rows

    rows = dict()
    for row in sheetdata:
        try:
            span = row["spans"].split(":")
        except:
            print(row)
            rows[row_idx] = []
            continue
        start = int(span[0])
        end = int(span[1])
        parsed_row = [None] * end 
        row_idx = None
        for cell in row:
            cell_name = cell["r"]
            col_str = "".join([c for c in cell_name if c.isalpha()])
            row_idx = int(cell_name[len(col_str):])
            
            column = get_index(col_str)
            
            is_string = "t" in cell.attrs
            is_formula = len(cell.contents) > 1

            if is_string:
                idx = int(cell.contents[0].contents[0])
                string = strings[idx]
                parsed_row[column] = string
            elif is_formula:
                formula = "=" + cell.contents[0].contents[0]
                parsed_row[column] = formula
            else:
                try:
                    val = int(cell.contents[0].contents[0])
                except:
                    val = float(cell.contents[0].contents[0])
                parsed_row[column] = val

        rows[row_idx] = parsed_row

    return rows

def get_cells(zf, strings):
    with zf.open('xl/worksheets/sheet1.xml') as myfile:
        soup = BeautifulSoup(myfile.read(), 'html5lib')
        sheetdata = soup.html.body.worksheet.dimension.sheetformatpr.sheetdata

        rows = get_rows(sheetdata, strings)

        if not any(rows):
            return []
        
        max_row = max(rows) 
        mdl = [[]] *  max_row
        for row in rows:
            mdl[row - 1] = rows[row]
            
    return mdl

def parse_excel_file(file_id, file_name, file_path):

    with ZipFile(file_path, 'r') as zf:
        # first we build the strings map
        strings = get_strings(zf)

        # then, we build the cells map
        cells = get_cells(zf, strings)

    return File(file_id, "excel", file_path, file_name, cells)

def write_excel_file(file):
    import xlsxwriter
    import os

    if os.path.exists(file.file_path):
        os.remove(file.file_path)

    workbook = xlsxwriter.Workbook(file.file_path)
    worksheet = workbook.add_worksheet()
    for row_idx, row in enumerate(file.file_contents.mixed_data_type):
        for col_idx, val in enumerate(row):
            worksheet.write(row_idx, col_idx, val)

    workbook.close()




import os
import pickle
import hashlib
import copy
import filecmp
import glob
import requests 
import os
import glob
import json
import shutil
from os.path import join, isfile
from saga.Commit import Commit
from saga.path_utils import copy_dir_to_dir, copy_file_to_dir, relative_paths_in_dir, changed_files
from saga.file_types.file_utils import parse_file, write_file


class Repository(object):

    def __init__(self, directory):
        self.base_directory = directory
        self.saga_directory = self.base_directory + "/.saga/"
        self.commit_directory = self.base_directory + "/.saga/commits/"
        self.state_directory = self.base_directory + "/.saga/states/"
        self.index_directory = self.base_directory + "/.saga/index/"
        self.repo_pickle = self.base_directory + "/.saga/repository"
        self.head = "master"
        self.branches = {"master" : None} # map from branch name -> commit hash
        self.index = {"master" : set()} # map from branch name -> set of files we are tracking
        self.file_ids = {"master": dict()} # map from branch -> file_name -> permanent_file_id
        self.commits = {"master": []} # this is an ugly hack for now, but we keep track of commit hashes on any branch

    @staticmethod
    def init(directory):
        # we make a new .saga repository, if possible
        if os.path.exists(directory + "/.saga"):
            raise ValueError("There is already a saga project at {}".format(directory))
        os.mkdir(directory + "/.saga")
        os.mkdir(directory + "/.saga/commits")
        os.mkdir(directory + "/.saga/states")
        os.mkdir(directory + "/.saga/index")
        repo = Repository(directory)
        repo.commit("Create repository", is_empty=True) # we add this empty commit so we don't need special cases
        return repo

    @staticmethod
    def read(directory):
        # read from the saga repo pickle
        repo_pickle = directory + "/.saga/repository"
        f = open(repo_pickle, "rb")
        p = f.read()
        f.close()
        return pickle.loads(p)

    def write(self):
        # read from the saga repo pickle
        p = pickle.dumps(self)
        f = open(self.repo_pickle, "wb+")
        f.write(p)
        f.close()

    def add(self, path):
        """
        If a file or folder exists at path, adds to the working index
        """
        if not os.path.exists(path):
            print("Error: path {} does not exist".format(path))
            return

        self._try_create_file_ids([path])

        if os.path.isfile(path):
            copy_file_to_dir(path, self.index_directory)
        else:
            assert os.path.isdir(path)
            files_in_path = relative_paths_in_dir(path)
            # we filter out the saga folder, if need be
            files_in_path = [join(path, f) for f in files_in_path if not f.startswith(".saga")]
            self._try_create_file_ids(files_in_path)
            # copy directory to index
            copy_dir_to_dir(path, join(self.index_directory, path), exclude=[".saga"])

    def commit(self, commit_message, is_empty=False):
        """
        Creates a new commit with commit_message, if there are new changes to commit
        """
        state_hash = self.index_hash()
        parent_commit_hash = self.branches[self.head]

        if not is_empty and not self._uncommited_in_index():
            print("Not commiting: no changes to commit")
            return

        commit = Commit(state_hash, [parent_commit_hash], commit_message)
        commit_hash = self._add_commit_to_db(commit)
        print("[{}] {}".format(commit_hash[0:12], commit_message))
        self.commits[self.head].append(commit_hash)
        self._backup_index_to_db()
        self.branches[self.head] = commit_hash

    def create_branch(self, branch_name):
        """
        If branch_name does not exist, will create a new branch with this name, 
        branching off from the current commit on head
        """
        if branch_name in self.branches:
            print("Error: branch {} already exists".format(branch_name))
        self.branches[branch_name] = self.branches[self.head]
        self.index[branch_name] = copy.deepcopy(self.index[self.head])
        self.file_ids[branch_name] = copy.deepcopy(self.file_ids[self.head])
        self.commits[branch_name] = copy.deepcopy(self.commits[self.head])

    def switch_to_branch(self, branch_name):
        """
        Switches the current head to branch name. Will fail if there are uncommited
        or unstaged changes (that aren't inserts).
        """
        if branch_name not in self.branches:
            print("Error: cannot switch to branch {} as it does not exist".format(branch_name))
            return

        if self._uncommited_in_index():
            print("Error: cannot switch branch as there are uncommited changed")
            return

        _, changed, removed = changed_files(self.base_directory, self.curr_state_dir())
        if any(changed) or any(removed):
            print("Error: please commit or undo changes before switching branches")
            return

        print("Switching to branch {}".format(branch_name))
        self.head = branch_name
        commit = self.get_commit(self.branches[self.head])
        self._restore_state(commit.state_hash)

    def log(self):
        """
        Prints all commits in the current branch to the screen
        """
        for commit_hash in reversed(self.commits[self.head]):
            commit = self.get_commit(commit_hash)
            print("commit: {}".format(commit_hash))
            print("\t{}\n".format(commit.commit_message))
        
    def status(self):
        """
        Prints the status of the current branch and index
        """
        print("On branch {}".format(self.head))

        rem_state_to_index, cha_state_to_index, ins_state_to_index = changed_files(self.curr_state_dir(), self.index_directory)
        rem_index_to_curr, cha_index_to_curr, ins_index_to_curr = changed_files(self.index_directory, self.base_directory)

        print("Changes staged for commit:")
        for path in rem_state_to_index:
            print("\tremoved: {}".format(path))
        for path in ins_state_to_index:
            print("\tinserted: {}".format(path))
        for path in cha_state_to_index:
            if path not in cha_index_to_curr and path not in rem_index_to_curr:
                print("\tmodified: {}".format(path))

        print("Change not staged for commit:")
        for path in rem_index_to_curr:
            print("\tremoved: {}".format(path))
        for path in cha_index_to_curr:
            print("\tmodified: {}".format(path))

        print("Untracked files:")
        for path in ins_index_to_curr:
            print("\t{}".format(path))

    def diff(self):
        """
        Prints the differences between all files
        """
        _, changed_paths, _ = changed_files(self.index_directory, self.base_directory)

        print("Git diff:")
        operations = []
        for path in changed_paths:
            if os.path.isfile(path):
                print("\tFile:", path)
                # we should read in the current file and the old file in this case
                file_id = self.file_ids[self.head][path]
                old_file = parse_file(file_id, path, join(self.index_directory, path))
                new_file = parse_file(file_id, path, join(self.base_directory, path))
                ops = old_file.get_operations(new_file)
                for key in ops:
                    print("\t\t:", key, ops[key])
                operations.extend(ops)
            else:
                print("\tDirectory:", path, "changed")

        return operations

    def branch(self):
        """
        Lists the branches in the repo, as well as the head
        """
        for branch in self.branches:
            if branch == self.head:
                print("   * {}".format(branch))
            else:
                print("     {}".format(branch))

    def merge(self, other_branch):
        """
        Merges current head with a other_branch
        """
        if other_branch not in self.branches:
            print("Error: cannot merge branch {} as it does not exist".format(other_branch))
            return


        lca, need_merge = self.least_common_ancestor(self.head, other_branch)

        if not need_merge:
            if len(self.commits[other_branch]) > len(self.commits[self.head]):
                self.branches[self.head] = self.branches[other_branch]
                self.commits[self.head] = copy.deepcopy(self.commits[other_branch])
                commit = self.get_commit(self.branches[self.head])
                self._restore_state(commit.state_hash)
                print("Fast-forward branch {} to branch {}".format(self.head, other_branch))
            else:
                print("Branch {} is a subbranch of {}".format(other_branch, self.head))
        else:

            state_dir_old = join(self.state_directory, self.get_commit(lca).state_hash)
            state_dir_newA = join(self.state_directory, self.state_hash())
            state_dir_newB = join(self.state_directory, self.get_commit(self.branches[other_branch]).state_hash)

            removed_paths_A, changed_paths_A, inserted_paths_A = changed_files(state_dir_old, state_dir_newA)
            removed_paths_B, changed_paths_B, inserted_paths_B = changed_files(state_dir_old, state_dir_newB)

            possibly_mergable = set() 
            print(inserted_paths_A)
            print(inserted_paths_B)
            for path in removed_paths_A.intersection(changed_paths_B):
                print("Confict as {} was removed in {} and modified in {}".format(path, self.head, other_branch))
            for path in removed_paths_B.intersection(changed_paths_A):
                print("Confict as {} was removed in {} and modified in {}".format(path, self.head, other_branch))
            for path in changed_paths_A.intersection(changed_paths_B):
                possibly_mergable.add(path)
                #print("Conflict as {} was modified in both {} and {}".format(path, self.head, other_branch))
            for path in inserted_paths_A.intersection(inserted_paths_B):
                print("Conflict as {} was inserted in both {} and {}".format(path, self.head, other_branch))

            files_to_write = [] # we only write the files if there are no merge conflicts in any of them
            for path in possibly_mergable:
                if path not in self.file_ids[self.head]: # we aren't tracking it
                    continue

                file_id = self.file_ids[self.head][path]
                file_o = parse_file(file_id, path, join(state_dir_old, path))
                file_a = parse_file(file_id, path, join(state_dir_newA, path))
                file_b = parse_file(file_id, path, join(state_dir_newB, path))

                merge_file = file_o.merge(file_a, file_b)
                if merge_file is None:
                    print("Merge conflict for file {}".format(file_o.file_name))
                    return
                else:
                    print("Successfully merged file {}".format(file_o.file_name))
                    files_to_write.append(merge_file)
            for f in files_to_write:
                write_file(f)
            self.commit("Merged {} into branch {}".format(other_branch, self.head))

    def pull(self):
        URL = "http://localhost:3000/cli/get-folder"

        shutil.rmtree(self.saga_directory)

        try:
            response = requests.get(url = URL, data={'folder_location' : ".saga"})
            paths = response.json()

            # create all the folders we need
            for path in paths["folder_paths"]:
                os.makedirs(os.path.join(self.base_directory, path))

            # and then download all the files in these folders
            for path in paths["file_paths"]:
                self._pull_file(path)
        except:
            print("Error: cannot pull")

    def push(self):
        relative_paths = relative_paths_in_dir(".saga")

        for path in relative_paths:
            abs_path = os.path.join(".saga", path)
            if os.path.isfile(abs_path):
                # Found a file 
                self._push_file(abs_path)
            else:
                self._push_folder(abs_path)

    def _pull_file(self, relative_file_path):
        # api-endpoint 
        URL = "http://localhost:3000/cli/download"

        # sending get request and saving the response as response object 
        r = requests.get(url = URL, data={'file_location' : relative_file_path})

        f = open(relative_file_path,'wb+')
        f.write(r.content)
        f.close()
        print("Downloaded File: {}".format(relative_file_path))

    def _push_file(self, relative_file_path):
        # api-endpoint 
        URL = "http://localhost:3000/cli/single-upload"

        file_name = os.path.basename(relative_file_path)

        with open(relative_file_path, 'rb') as f:
            r = requests.post(URL, files={"file": f}, data={"relative_file_path": relative_file_path, "file_name": file_name})

    def _push_folder(self, folder_path):
        # api-endpoint 
        URL = "http://localhost:3000/cli/push-folder"

        requests.post(url=URL, data={"folder_path": folder_path})

    def state_hash(self):
        return self.get_commit(self.branches[self.head]).state_hash

    def curr_state_dir(self):
        return join(self.state_directory, self.get_commit(self.branches[self.head]).state_hash)

    def least_common_ancestor(self, branch_1, branch_2):
        branch_1_commits = self.commits[branch_1]
        branch_2_commits = self.commits[branch_2]

        for i in reversed(range(min(len(branch_1_commits), len(branch_2_commits)))):
            if branch_1_commits[i] == branch_2_commits[i]:
                if i == min(len(branch_1_commits), len(branch_2_commits)) - 1:
                    return branch_1_commits[i], False
                else:
                    return branch_1_commits[i], True 
        
        raise Exception("No common ancestor")

    def _backup_index_to_db(self):
        # copies all files that are being tracked to a state hash directory
        dst = join(self.state_directory, self.index_hash())
        if not os.path.isdir(dst):
            os.mkdir(dst)
            # TODO: make this work when the cwd isn't the base repository
            copy_dir_to_dir(".saga/index/", dst)

    def _restore_state(self, state_hash):
        copy_dir_to_dir(join(".saga/states", state_hash), ".")
    
    def get_commit(self, commit_hash):
        f = open(self.commit_directory + commit_hash, "rb")
        commit_bytes = f.read()
        f.close()
        return Commit.from_bytes(commit_bytes)

    def _add_commit_to_db(self, commit):
        # if the commit exists, we don't need to error
        commit_hash = commit.hash()
        commit_bytes = commit.to_bytes()
        f = open(self.commit_directory + commit_hash, "wb+")
        f.write(commit_bytes)
        f.close()
        return commit_hash

    def index_hash(self):
        file_hashes = []
        # we sort to have the same files every time
        for file_name in sorted(relative_paths_in_dir(self.index_directory)):
            index_file_name = join(self.index_directory, file_name)
            if os.path.isfile(index_file_name):
                f = open(index_file_name, "rb")
                file_bytes = f.read()
                m = hashlib.sha256()
                m.update(file_bytes) # we encode the file contents
                m.update(file_name.encode("utf-8")) # and we encode the file path
                file_hashes.append(m.hexdigest())
        m = hashlib.sha256()
        #TODO: this should be a merkle tree eventually 
        # (or maybe an merkel-mountain-range), to reap the benefits
        m.update(",".join(file_hashes).encode('utf-8')) 
        return m.hexdigest()

    def get_new_file_id(self):
        # generates a random ID for a file
        import uuid
        return uuid.uuid4().hex

    def _try_create_file_ids(self, file_paths):
        for path in file_paths:
            if path not in self.file_ids[self.head]:
                self.file_ids[self.head][path] = self.get_new_file_id()

    def _uncommited_in_index(self):
        index_state_hash = self.index_hash()
        commit_state_hash = self.state_hash()
        return index_state_hash != commit_state_hash

    def restore_state_to_head(self):
        self._restore_state(self.state_hash())


import os
from saga.file_types.file_utils import parse_file


def do_test_merge(folder_path, ending):
    """
    folder_path should be a path to a folder with:
        - an original version of a file named `origin`
        - two modified versions, `a` and `b`
        - a merged version of these changes in the file `merge`

    This returns true iff a and b merged with the original as base
    results in the merged file!
    """ 
    origin = parse_file("id", "file", os.path.join(folder_path, "origin" + ending))
    a = parse_file("id", "file", os.path.join(folder_path, "a" + ending))
    b = parse_file("id", "file", os.path.join(folder_path, "b" + ending))
    merge = parse_file("id", "file", os.path.join(folder_path, "merge" + ending))

    calculated_merge = origin.merge(a, b)

    return merge.file_contents == calculated_merge.file_contents






import pytest
from saga.base_file.mixed_data_type.diff_utils import inserted_paths, changed_paths, removed_paths
from saga.base_file.mixed_data_type.lcs import lcs_multi_dimension

def check_ops(A, B, num_removed, num_changed, num_inserted):
    removed = removed_paths(A, B)
    changed = changed_paths(A, B)
    inserted = inserted_paths(A, B)

    assert len(removed) == num_removed
    assert len(changed) == num_changed
    assert len(inserted) == num_inserted

def test_list_add_element_end():
    check_ops(["A", "B"], ["A", "B", "C"], 0, 0, 1)

def test_list_add_element_middle():
    check_ops(["A", "C"], ["A", "B", "C"], 0, 0, 1)

def test_list_add_element_start():
    check_ops(["B", "C"], ["A", "B", "C"], 0, 0, 1)

def test_list_add_elements():
    check_ops([], ["A", "B", "C"], 0, 0, 3)

def test_list_rem_element_end():
    check_ops(["A", "B", "C"], ["A", "B"], 1, 0, 0)

def test_list_rem_element_middle():
    check_ops(["A", "B", "C"], ["A", "C"], 1, 0, 0)

def test_list_rem_element_start():
    check_ops(["A", "B", "C"], ["B", "C"], 1, 0, 0)

def test_list_rem_elements():
    check_ops(["A", "B", "C"], [], 3, 0, 0)

def test_list_change_elements():
    check_ops(["AB", "B", "C"], ["A", "B", "C"], 0, 1, 0)

def test_list_all_operations():
    check_ops(["AB", "B", "C"], ["A", "C", "D"], 1, 1, 1)

def test_list_in_list():
    check_ops([[]], [[], []], 0, 0, 1)
    check_ops([[]], [], 1, 0, 0)
    check_ops([["A"]], [["A", "B"]], 0, 0, 1)

def test_list_in_list_change():
    check_ops([[]], [[], []], 0, 0, 1)
    check_ops([[]], [], 1, 0, 0)
    check_ops([["A"]], [["ABC"]], 0, 1, 0)

def test_dict_in_dict():
    check_ops({"key":{"key": "value"}}, {"key":{"key": "value", "key1": "value"}}, 0, 0, 1)
    check_ops({"key":{"key": "value"}}, {}, 1, 0, 0)
    check_ops({"key":{"key": "value"}}, {"key":{"key": "value1"}}, 0, 1, 0)

def test_list_in_dict():
    pass

def test_dict_in_list():
    pass
    
    



import pytest
from saga.base_file.mixed_data_type.merge_utils import merge_rec


def test_list_simple_conflict():
    O = [1]
    A = [1, 2]
    B = [1, 3]

    result = merge_rec(O, A, B)
    assert result == None

def test_list_simple_merge():
    O = [1]
    A = [0, 1]
    B = [1, 2]

    result = merge_rec(O, A, B)
    assert result == [0, 1, 2]

def test_dict_simple_conflict():
    A = {"key": "value"}
    O = {"key": "value1"}
    B = {"key": "value2"}

    result = merge_rec(O, A, B)
    assert result == None

def test_dict_simple_merge():
    O = {"key": "value"}
    A = {"key": "value1"}
    B = {"key": "value"}

    result = merge_rec(O, A, B)
    assert result == {"key": "value1"}


def test_list_complicated_conflict():
    O = [1, 2, 3, 4, 5, 6]
    A = [1, 4, 5, 2, 3, 6]
    B = [1, 2, 4, 5, 3, 6]

    result = merge_rec(O, A, B)
    assert result == None


def test_list_successful_merge():
    O = [1, 2, 6]
    A = [1, 4, 5, 2, 6]
    B = [1, 2, 6]

    result = merge_rec(O, A, B)
    assert result == [1, 4, 5, 2, 6]


def test_multi_dim_merge_easy():
    O = [[1, 2], [3, 4]]
    A = [[1, 2, 3], [3, 4]]
    B = [[1, 2], [3, 4]]

    result = merge_rec(O, A, B)
    assert result == [[1, 2, 3], [3, 4]]


def test_multi_dim_merge_harder():
    O = [[1, 2], [6, 7], [3, 4]]
    A = [[1, 2, 3], [6, 7], [3, 4]]
    B = [[1, 2], [6, 7], [3, 4, 5]]

    result = merge_rec(O, A, B)
    assert result == [[1, 2, 3], [6, 7], [3, 4, 5]]

def test_multi_dim_merge_recursive():
    O = [[1, 2], [6, 7], [3, 4]]
    A = [[1, 2, 3], [6, 7], [3, 4]]
    B = [[0, 1, 2], [6, 7]]

    result = merge_rec(O, A, B)
    assert result == [[0, 1, 2, 3], [6, 7]]


def test_multi_dim_merge_recursive_add_on_end():
    O = [[1, 2], [6, 7]]
    A = [[1, 2, 3], [6, 7], [3, 4]]
    B = [[0, 1, 2], [6, 7], [3, 4]]

    result = merge_rec(O, A, B)
    assert result == None

def test_easy_multi_dim_dict():
    O = {"key": {"key": "value"}}
    A = {"key": {"key": "value1"}}
    B = {"key": {"key": "value"}}

    result = merge_rec(O, A, B)
    assert result == {"key": {"key": "value1"}}

def test_conflict_multi_dim_dict():
    O = {"key": {"key": "value"}}
    A = {"key": {"key": "value1"}}
    B = {"key": {"key": "value2"}}

    result = merge_rec(O, A, B)
    assert result == None

def test_harder_multi_dim_dict():
    O = {"key": {"key": "value"}, "key1": {"key": "value"}}
    A = {"key": {"key": "value"}}
    B = {"key": {"key": "value2"}}

    result = merge_rec(O, A, B)
    assert result == {"key": {"key": "value2"}}

def test_list_of_dicts():
    O = [{"key": "value"}, {"key1": "value"}, {"key2": "value"}]
    A = [{"key": "value1"}, {"key1": "value"}, {"key2": "value"}]
    B = [{"key": "value"}, {"key1": "value1"}, {"key2": "value"}]

    result = merge_rec(O, A, B)
    assert result == [{"key": "value1"}, {"key1": "value1"}, {"key2": "value"}]


def test_dict_of_lists():
    O = {"key": [1]}
    A = {"key": [0, 1]}
    B = {"key": [1, 2]}

    result = merge_rec(O, A, B)
    assert result == {"key": [0, 1, 2]}

def test_dict_of_lists_of_dicts():
    O = {"key": [{"key": "value"}, {"key1": "value"}, {"key2": "value"}]}
    A = {"key": [{"key": "value1"}, {"key1": "value"}, {"key2": "value"}]}
    B = {"key": [{"key": "value"}, {"key1": "value1"}, {"key2": "value"}]}

    result = merge_rec(O, A, B)
    assert result == {"key": [{"key": "value1"}, {"key1": "value1"}, {"key2": "value"}]}


import pytest
import os
from saga.base_file.File import File
from saga.file_types.text_file import parse_text_file, write_text_file

def create_text_file(path, contents):
    if os.path.isfile(path):
        os.remove(path)
    print("path")
    f = open(path, 'w+')
    f.write(contents)
    f.close()

@pytest.fixture()
def setup_text_files():
    if not os.path.isdir("temp"):
        os.mkdir('temp')
    create_text_file("temp/text0", "line1\nline2")

def test_create(setup_text_files):
    text_file = parse_text_file("id", "name", "temp/text0")
    assert text_file.file_name == "name"
    assert text_file.file_contents.mixed_data_type == ["line1", "line2"]

def test_create_then_write(setup_text_files):
    text_file = parse_text_file("id", "name", "temp/text0")
    write_text_file(text_file)
    f = open('temp/text0', 'r')
    assert f.read() == "line1\nline2"

def test_get_operations_no_change(setup_text_files):
    text_file0 = parse_text_file("id", "name", "temp/text0")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)
    for key in ops:
        assert len(ops[key]) == 0
    
def test_get_operations_insert_end(setup_text_files):
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "line1\nline2\nline3")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)
    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 1
    assert ops["inserted"] == [[2]]
    assert len(ops["changed"]) == 0
    assert len(ops["removed"]) == 0

def test_get_operations_insert_middle(setup_text_files):
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "line1\nline3\nline2")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)
    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 1
    assert ops["inserted"] == [[1]]
    assert len(ops["changed"]) == 0
    assert len(ops["removed"]) == 0


def test_get_operations_insert_multiple_middle(setup_text_files):
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "line1\nline3\nline4\nline2")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)
    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 2
    assert ops["inserted"] == [[1], [2]]
    assert len(ops["changed"]) == 0
    assert len(ops["removed"]) == 0

def test_get_operations_remove(setup_text_files):
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "line1")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)
    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 0
    assert len(ops["changed"]) == 0
    assert len(ops["removed"]) == 1
    assert ops["removed"] == [[1]]

def test_get_operations_remove_all(setup_text_files):
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)

    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 0
    assert len(ops["changed"]) == 0
    assert len(ops["removed"]) == 2
    assert ops["removed"] == [[0], [1]]

def test_get_operations_remove_mulitple(setup_text_files):
    create_text_file("temp/text0", "1\n2\n3\n4")
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "1\n4\n")
    text_file1 = parse_text_file("id", "name", "temp/text0")
    ops = text_file0.get_operations(text_file1)
    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 0
    assert len(ops["changed"]) == 0
    assert len(ops["removed"]) == 2
    assert ops["removed"] == [[1], [2]]

def test_get_operations_insert_remove_change(setup_text_files):
    create_text_file("temp/text0", "line1\nline2111\nline3\nline4")
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", "line1\nline211\nline4\nxyz")
    text_file1 = parse_text_file("id", "name", "temp/text0")

    ops = text_file0.get_operations(text_file1)
    
    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 1
    assert ops["inserted"] == [[3]]
    assert len(ops["changed"]) == 1
    assert ops["changed"] == [[1]]
    assert len(ops["removed"]) == 1
    assert ops["removed"] == [[2]]

def test_complex_operations(setup_text_files):
    old_code_list = [
        "for (int i = 0; i < 100; i++ ) {",
        "System.out.println(i * 2);",
        "int temp = i * 1000;",
        "if (temp < i) {",
        "System.out.println(temp, i);",
        "}",
        "}",
    ]
    old_code = "\n".join(old_code_list)
    new_code_list = [
        "for (int j = 0; j < 100; j++ ) {",
        "int temp1 = j * 1000;",
        "if (temp1 < j) {",
        "System.out.println(temp1, j);",
        "}",
        "}",
    ]
    new_code = "\n".join(new_code_list)

    create_text_file("temp/text0", old_code)
    text_file0 = parse_text_file("id", "name", "temp/text0")
    create_text_file("temp/text0", new_code)
    text_file1 = parse_text_file("id", "name", "temp/text0")

    ops = text_file0.get_operations(text_file1)

    assert len(ops["file"]) == 0
    assert len(ops["inserted"]) == 0
    assert len(ops["changed"]) == 4
    assert ops["changed"] == [[0], [2], [3], [4]]
    assert len(ops["removed"]) == 1
    assert ops["removed"] == [[1]]


def test_merge():
    from tests.test_utils import do_test_merge
    for path in os.listdir("tests/merges/text"):
        path = os.path.join("tests/merges/text", path)
        if os.path.isdir(path):
            assert do_test_merge(path, ".txt")


import pickle
import hashlib

class State():

    def __init__(self, files, prev_state_hash):
        # map from file_id -> file object
        self.files = files 
        self.prev_state_hash = prev_state_hash

    def __getitem__(self, file_id):
        return self.files.get(file_id, None) # we return None of the file does not exist

    def __setitem__(self, file_id, file):
        self.files[file_id] = file

    def __contains__(self, file_name):
        return file_name in self.files

    def __len__(self):
        return sum(1 for file_id in self.files if self.files[file_id] is not None)

    def get_hash(self):
        m = hashlib.sha256()
        m.update(bytearray(self.to_string()))
        return m.hexdigest()

    def to_string(self):
        return pickle.dumps(self)

    def from_string(self, str):
        return pickle.loads(str)

    


    



"""
Functions for computing longest-common subsequences between lists
"""
from copy import deepcopy

# LCS with a similarity function. 
def lcs_similarity(A, B, similarity_function):
    m = len(A) 
    n = len(B) 

    # L[i][j] is the highest similar metric of the close longest common subsequece up to indexes i, j
    L = [[None]*(n + 1) for i in range(m + 1)] 
    for i in range(m + 1): 
        for j in range(n + 1): 
            if i == 0 or j == 0 : 
                L[i][j] = 0
            else:
                L[i][j] = max(
                    similarity_function(A[i - 1], B[j - 1]) + L[i - 1][j - 1],
                    L[i - 1][j],
                    L[i][j - 1]
                )

    # Create an array to store indexes of close common subsequence
    matches = []

    # Start from the right-most-bottom-most corner and 
    # one by one store characters in lcs[] 
    i = m 
    j = n 
    while i > 0 and j > 0:

        # If current character in X[] and Y are same, then 
        # current character is part of LCS 
        if similarity_function(A[i - 1], B[j - 1]) + L[i - 1][j - 1] > max(L[i - 1][j], L[i][j - 1]): 
            matches.append(([i - 1], [j - 1], similarity_function(A[i - 1], B[j - 1])))
            i-=1
            j-=1

        # If not same, then find the larger of two and 
        # go in the direction of larger value 
        elif L[i-1][j] > L[i][j-1]:
            i-=1
        else: 
            j-=1
    
    matches.reverse()

    return matches

# Standard LCS: the similarity function is just exact equality
def lcs(A, B):
    def equal(a, b):
        if a == b:
            return 1
        return 0

    return lcs_similarity(A, B, equal)


def list_from_path(matrix, path):
    curr = matrix
    for step in path:
        curr = curr[step]
    return curr

def same_paths(dict_a, dict_b):
    return {k for k in dict_a if k in dict_b and dict_a[k] == dict_b[k]}

def similarity_function(A, B):
    if type(A) != type(B):
        return 0
    elif A is None:
        return 1
    elif type(A) in (int, float, bool):
        return 1 if A == B else 0
    elif type(A) == str:
        if len(A) == 0 and len(B) == 0:
            return 1
        indexes = lcs(A, B)
        if any(indexes):
            return len(indexes) / max(len(A), len(B))
        return 0
    elif type(A) == list:
        if len(A) == 0 and len(B) == 0:
            return 1
            
        A = [a for a in A if a is not None]
        B = [b for b in B if b is not None]

        indexes = lcs_similarity(A, B, similarity_function)
        if any(indexes):
            total = sum(sim for _, _, sim in indexes)
            return total / max(len(A), len(B))
        return 0
    elif type(A) == dict:
        # we just do the number of unchanged keys divided by the max number of keys
        unchanged = same_paths(A, B)
        if any(unchanged):
            return len(unchanged) / max(len(A), len(B))
        return 0

    else:
        print("Unknown type {}".format(type(A)))
        return 0

def lcs_with_sim(A, B):
    return lcs_similarity(A, B, similarity_function)

# returns a mapping from "dimension down" to "matches at that level"
def lcs_multi_dimension(A, B):    
    dimension_matches = dict()
    dimension_matches[1] = lcs_similarity(A, B, similarity_function)

    dimension = 1
    matches = dimension_matches[1]
    while any(matches):
        dimension += 1
        dimension_matches[dimension] = []
        for path_a, path_b, _ in dimension_matches[dimension - 1]:
            list_a = list_from_path(A, path_a)
            list_b = list_from_path(B, path_b)
            if type(list_a) != list or type(list_b) != list:
                matches = []
                break

            matches = lcs_similarity(list_a, list_b, similarity_function)
            for idx_a, idx_b, sim in matches:
                dimension_matches[dimension].append((path_a + idx_a, path_b + idx_b, sim))

    return dimension_matches

def get_matching_path(dim_matches, path_is_A, path):
    for path_a, path_b, sim in dim_matches[len(path)]:
        if path_is_A:
            if path_a == path:
                return path_b, sim
        else:
            if path_b == path:
                return path_a, sim
    return None, None



    




import os

def test_merge():
    from tests.test_utils import do_test_merge
    for path in os.listdir("tests/merges/excel"):
        path = os.path.join("tests/merges/excel", path)
        if os.path.isdir(path):
            assert do_test_merge(path, ".xlsx")


import os
import shutil
import glob
import filecmp

def copy_dir_to_dir(src, dst, exclude=None):
    """
    src is a relative path. It must be a folder that exists.
    dst is an absolute path. It will be created if it does not exist.

    If a path = src/file_name, then this will be copied to dst/file_name
    """
    if exclude is None:
        exclude = []
    if src[-1] == "/":
        src = src[:-1]

    if not os.path.exists(dst):
            os.makedirs(dst)

    # otherwise, we recursively expore the directory and copy it over
    for root, dirs, files in os.walk(src):
        dirs[:] = [d for d in dirs if d not in exclude]
        relative_root = root[len(src) + 1:]

        # first we copy the directories
        for directory in dirs:
            # if the directory doesn't exist, we make it
            path = os.path.join(dst, relative_root, directory)
            if not os.path.isdir(path):
                os.mkdir(path)

        # and then the files
        for file_name in files:
            path = os.path.join(dst, relative_root, file_name)
            shutil.copyfile(os.path.join(root, file_name), path)

def copy_file_to_dir(src, dst):
    """
    src is a relative path of a file. 
    dst is an absolute path of a folder. 

    src will eixst at join(dst, src)
    """
    dirname = os.path.join(dst, os.path.dirname(src))
    print(dirname)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    shutil.copyfile(src, os.path.join(dst, src))

def relative_paths_in_dir(directory, ignore=None):
    """
    Returns a list of paths in the directory, excluding paths that begin with ignore
    """
    if ignore is None:
        ignore = []

    # we make sure that the path does not start with a slash
    if directory[-1] == "/":
        paths = [path[len(directory):] for path in glob.glob(directory + '/**', recursive=True) if path[len(directory):] != ""]
    else:
        paths = [path[len(directory) + 1:] for path in glob.glob(directory + '/**', recursive=True) if path[len(directory) + 1:] != ""]
    def include_path(path):
        for ignore_path in ignore:
            if path.startswith(ignore_path):
                return False
        if path == directory:
            return False
        return True
    return {path for path in paths if include_path(path)}

def changed_files(dir1, dir2):
        previous_state = relative_paths_in_dir(dir1)
        current_state = relative_paths_in_dir(dir2)

        removed_paths, changed_paths, inserted_paths  = set(), set(), set()
        for path in previous_state:
            if path not in current_state:
                removed_paths.add(path)
            else:
                path1 = os.path.join(dir1, path)
                path2 = os.path.join(dir2, path)
                if os.path.isfile(path1) and os.path.isfile(path2) and not filecmp.cmp(path1, path2):
                    changed_paths.add(path)
                elif (os.path.isfile(path1) and not os.path.isfile(path2)) or \
                    (not os.path.isfile(path1) and os.path.isfile(path2)):
                    changed_paths.add(path)
        for path in current_state:
            if path not in previous_state:
                inserted_paths.add(path)

        return removed_paths, changed_paths, inserted_paths


import pickle
import hashlib

class Commit(object):

    def __init__(self, state_hash, parent_commit_hashes, commit_message):
        self.state_hash = state_hash
        self.parent_commit_hashes = parent_commit_hashes
        self.commit_message = commit_message

    def hash(self):
        m = hashlib.sha256()
        m.update(self.to_bytes())
        return m.hexdigest()
    
    def to_bytes(self):
        return pickle.dumps(self)

    @staticmethod
    def from_bytes(commit_bytes):
        return pickle.loads(commit_bytes)





import pytest
import os
from saga.base_file.File import File
from saga.file_types.csv_file import parse_csv_file, write_csv_file

def create_csv_file(name, contents):
    if os.path.isfile(name):
        os.remove(name)
    f = open(name, 'w+')
    f.write(contents)
    f.close()

@pytest.fixture()
def setup_csv_files():
    if not os.path.isdir("temp"):
        os.mkdir('temp')
    create_csv_file("temp/csv0", "name,email\ntest,test@gmail.com")

def test_create(setup_csv_files):
    csv_file = parse_csv_file("id", "name", "temp/csv0")
    assert csv_file.file_name == "name"
    assert csv_file.file_contents.mixed_data_type == [["name", "email"], ["test", "test@gmail.com"]]

def test_create_then_write(setup_csv_files):
    csv_file = parse_csv_file("id", "name", "temp/csv0")
    write_csv_file(csv_file)
    f = open('temp/csv0', 'r')
    assert f.read() == "name,email\ntest,test@gmail.com"

def test_get_operations_no_change(setup_csv_files):
    csv_file0 = parse_csv_file("id", "name", "temp/csv0")
    csv_file1 = parse_csv_file("id", "name", "temp/csv0")
    ops = csv_file0.get_operations(csv_file1)
    for key in ops:
        assert len(ops[key]) == 0
    
def test_get_operations_insert_row(setup_csv_files):
    csv_file0 = parse_csv_file("id", "name", "temp/csv0")
    create_csv_file("temp/csv0", "name,email\ntest,test@gmail.com\nharry,harry@gmail.com")
    csv_file1 = parse_csv_file("id", "name", "temp/csv0")
    ops = csv_file0.get_operations(csv_file1)
    assert len(ops["file"]) == 0
    assert len(ops["changed"]) == 0
    assert len(ops["inserted"]) == 1
    assert ops["inserted"] == [[2]]
    assert len(ops["removed"]) == 0
